package ssw.mj.test;

// TODO: move sym tab & code gen testing to another class, Goal: seperate test class for each assignment

import java.io.StringReader;

import ssw.mj.Messages;
import ssw.mj.Parser;
import ssw.mj.Scanner;
import ssw.mj.Token;
import ssw.mj.codegen.Code;
import ssw.mj.symtab.Obj;
import ssw.mj.symtab.Struct;
import ssw.mj.symtab.Tab;

/**
 * Test cases for the <code>Parser</code> class.
 *
 * @author AW
 */
public class ParserTest extends IdentifiedTestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";

	private static TestPrintWriter output;
	private static TestByteArrayOutputStream code;

	//---------- helpers ----------

	private void initParser (String program) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(program), null);
		code = new TestByteArrayOutputStream();
	}

	//--------------------------------
	//---------- test cases ----------

	/** Test the shortest syntactically valid program: */
	public void testShortestProgram () {
		initParser("program Shortest { }");

		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 20: main", "NOT_FOUND", "");
		output.verify();

		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		// just to make sure that the Universe is also there
		actual = Tab.find("null");
		expected = new Obj(Obj.Con, "null", Tab.nullType);
		expected.val = 0;
		assertEquals("Universe 'null'", expected, actual);

		// only name in symbol table of this program
		actual = Tab.find("Shortest");
		expected = new Obj(Obj.Prog, "Shortest", Tab.noType);
		expected.locals = null;
		assertEquals("Program 'Shortest'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);

		//////////////////////////
		// check generated code //
		//////////////////////////
		code.setCodeSize(0);
		Code.write(code);
		code.verify();
	}

	public void testDoNothingProgram () {
		initParser(	"program DoNothing" + LF +
					"\tint dummy;" + LF +
					"{" + LF +
					"\tint doNothing() { dummy=0; return dummy; }" + LF +
					"" + LF +
					"\tvoid main() { " + LF +
					"\t\tdummy= doNothing();" + LF +
					"\t}" + LF +
					"}");

		Parser.parse(output);
		output.verify();

		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		actual = Tab.find("DoNothing");
		expected = new Obj(Obj.Prog, "DoNothing", Tab.noType);
		// we cheat here, because we will compare all the rest later
		expected.locals = actual.locals;
		assertEquals("Program 'DoNothing'", expected, actual);

		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately
		// NOTE: thus we get better error messages

		actual = actual.locals;
		expected = new Obj(Obj.Var, "dummy", Tab.intType);
		expected.adr = 0;
		expected.level = 0;
		assertEquals("Global 'dummy'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Meth, "doNothing", Tab.intType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'doNothing'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);

		//////////////////////////
		// check generated code //
		//////////////////////////
				// dummy = 0;
		code.expect(15, 1);	// const_0
		code.expect(12, 1);	// putstatic
		code.expect(0, 2);	// s
				// return dummy;
		code.expect(11, 1);	// getstatic
		code.expect(0, 2);	// s
				// dummy = doNothing();
		code.expect(12, 1);	// putstatic
		code.expect(0, 2);	// s

		Code.write(code);

		code.setCodeSize(10);
		code.setDataSize(1);
		code.setMainPC(-1);
		code.verify();
	}

	public void testDoWhileSimple () {
		initParser("program DW {" + LF +
		           "  void main () int x; {" + LF +
		           "    do x = x + 1;" + LF +
		           "    while (x < 5);" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
				// x = x + 1;
		code.expect( 2, 1);		// load_0
		code.expect(16, 1);		// const_1
		code.expect(23, 1);		// add
		code.expect( 7, 1);		// store_0

		Code.write(code);

		code.setCodeSize(4);
		code.setDataSize(0);
		code.setMainPC(-1);
		code.verify();
	}

	public void testDoWhile_1And () {
		initParser("program DW {" + LF +
		           "  void main () int x; {" + LF +
		           "    do x = x + 1;" + LF +
		           "    while (x < 5 && x > 1);" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
				// x = x + 1;
		code.expect( 2, 1);		// load_0
		code.expect(16, 1);		// const_1
		code.expect(23, 1);		// add
		code.expect( 7, 1);		// store_0

		Code.write(code);

		code.setCodeSize(4);
		code.setDataSize(0);
		code.setMainPC(-1);
		code.verify();
	}

	public void testDoWhile_1Or_2And () {
		initParser(
			"program DW {" + LF +
			"  void main () int x; {" + LF +
			"    do x = x + 1;" + LF +
			"    while (x < 5 && x > 1 || x != 3 && x != 2);" + LF +
			"  }" + LF +
			"}"
		);

		Parser.parse(output);
		output.verify();

		//////////////////////////
		// check generated code //
		//////////////////////////
				// x = x + 1;
		code.expect(2, 1);		// load_0
		code.expect(16, 1);		// const_1
		code.expect(23, 1);		// add
		code.expect(7, 1);		// store_0

		Code.write(code);

		code.setCodeSize(4);
		code.setDataSize(0);
		code.setMainPC(-1);
		code.verify();
	}

	public void testRefTypeComparison () {
		initParser(
			"program RTC" + LF +
			"  int[] iarr, jarr;" + LF +
			"{" + LF +
			"  void test () {" + LF +
			"    if (iarr == null);" + LF +
			"    if (iarr != null);" + LF +
			"    if (iarr == jarr);" + LF +
			"    if (iarr != jarr);" + LF +
			"  }" + LF +
			"  void main () { }" + LF +
			"}"
		);

		Parser.parse(output);
		output.verify();
	}

	/* suggested by KBP */
	public void testTwoReturns () {
		initParser("program TR {" + LF +
		           "  void main () {}" + LF +
		           "  int getInt (int x) {" + LF +
		           "    if (x > 5) return 0;" + LF +
		           "    if (x > 0) return 1;" + LF +
		           "    return 2;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.verify();
	}

	/* suggest by KBP */
	public void testNestedLoops () {
		initParser(
				"program NL {" + LF +
				"  void main () int i; {" + LF +
				"    i = 0;" + LF +
				"    while (i < 10) {" + LF +
				"      while (i < 5) {" + LF +
				"        i++;" + LF +
				"        if (i == 3) break;" + LF +
				"      }" + LF +
				"      i++;" + LF +
				"    }" + LF +
				"  }" + LF +
				"}"
		);

		Parser.parse(output);
		output.verify();
		
		// TODO: test generated code
	}

	/* suggest by KBP */
	public void testFuncAsProc () {
		initParser("program FAP {" + LF +
							 "  int getInt () { return 1; }" + LF +
							 "  void main () {" + LF +
							 "    getInt();" + LF +
							 "  }" + LF +
							 "}");

		Parser.parse(output);
		output.verify();
		
		// TODO: test generated code
	}

	public void testScriptExample () {
		initParser( "program P" + LF +
		            "" + LF +
		            "  final int size = 10;" + LF +
		            "" + LF +
		/*line  5*/ "  class Table {" + LF +
		            "    int[] pos;" + LF +
		            "    int[] neg;" + LF +
		            "  }" + LF +
		            "" + LF +
		/*line 10*/ "  Table val;" + LF +
		            "{" + LF +
		            "  void main()" + LF +
		            "    int x, i;" + LF +
		            "  { /*---------- Initialize val */" + LF +
		/*line 15*/ "    val = new Table;" + LF +
		            "    val.pos = new int[size]; val.neg = new int[size];" + LF +
		            "    i = 0;" + LF +
		            "    while (i < size) {" + LF +
		            "      val.pos[i] = 0; val.neg[i] = 0;" + LF +
		/*line 20*/ "      i++;" + LF +
		            "    }" + LF +
		            "    /*---------- Read values */" + LF +
		            "    read(x);" + LF +
		            "    while (x > 0) {" + LF +
		/*line 25*/ "      if (0 <= x && x < size) {" + LF +
		            "        val.pos[x]++;" + LF +
		            "      } else if (-size < x && x < 0) {" + LF +
		            "        val.neg[-x]++;" + LF +
		            "      }" + LF +
		/*line 30*/ "      read(x);" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "}");

		Parser.parse(output);
		output.verify();

		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected, locals;
		Struct tableT;

		actual = Tab.find("P");  // class P
		expected = new Obj(Obj.Prog, "P", Tab.noType);
		// We cheat here, because we will compare all the rest later.
		expected.locals = actual.locals;
		assertEquals("Program 'P'", expected, actual);

		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately.
		// NOTE: This gives us more precise error messages
		//       than just re-creating the whole symbol table and
		//       have it check by one call to equals(Prog).

		actual = actual.locals;  // final int size = 10;
		expected = new Obj(Obj.Con, "size", Tab.intType);
		expected.val = 10;
		assertEquals("Constant 'size'", expected, actual);

		actual = actual.next;  // class Table { int[] pos; int[] neg; }
		tableT = new Struct(Struct.Class);
		tableT.n = 2;
		// No cheating here, field list is small enough.
		// So we explicitly re-create it and check it at once (with equals).
		expected = new Obj(Obj.Var, "pos", new Struct(Struct.Arr, Tab.intType));
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Field 'pos'", expected, actual.type.fields);
		tableT.fields = expected;
		expected = new Obj(Obj.Var, "neg", new Struct(Struct.Arr, Tab.intType));
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Field 'neg'", expected, actual.type.fields.next);
		tableT.fields.next = expected;
		expected = new Obj(Obj.Type, "Table", tableT);
		assertEquals("Type 'Table'", expected, actual);

		actual = actual.next;  // Table val;
		expected = new Obj(Obj.Var, "val", tableT);
		assertEquals("Global 'val'", expected, actual);

		actual = actual.next;  // void main () int x, i;
		expected = new Obj(Obj.Var, "x", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Local 'x'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Var, "i", Tab.intType);
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Local 'i'", expected, actual.locals.next);
		locals.next = expected;
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = locals;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);

		//////////////////////////
		// check generated code //
		//////////////////////////
				// val = new Table;						(6 byte)
		code.expect(32, 1);		// new
		code.expect(8, 2);		// s
		code.expect(12, 1);		// putstatic
		code.expect(0, 2);		// s
				// val.pos = new int[size];				(13 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(22, 1);		// const_
		code.expect(10, 4);		// w
		code.expect(33, 1);		// newarray
		code.expect(1, 1);		// b
		code.expect(14, 1);		// putfield
		code.expect(0, 2);		// s
				// val.neg = new int[size];				(13 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(22, 1);		// const_
		code.expect(10, 4);		// w
		code.expect(33, 1);		// newarray
		code.expect(1, 1);		// b
		code.expect(14, 1);		// putfield
		code.expect(1, 2);		// s
				// i = 0;								(2 byte)
		code.expect(15, 1);		// const_0
		code.expect(8, 1);		// store_1
					// val.pos[i] = 0;					(9 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(13, 1);		// getfield
		code.expect(0, 2);		// s
		code.expect(3, 1);		// load_1
		code.expect(15, 1);		// const_0
		code.expect(35, 1);		// astore
					// val.neg[i] = 0;					(9 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(13, 1);		// getfield
		code.expect(1, 2);		// s
		code.expect(3, 1);		// load_1
		code.expect(15, 1);		// const_0
		code.expect(35, 1);		// astore
					// i++;								(3 byte)
		code.expect(31, 1);		// inc
		code.expect(1, 1);		// b1
		code.expect(1, 1);		// b2
				// read(x);								(2 byte)
		code.expect(53, 1);		// read
		code.expect(7, 1);		// store_0
						// val.pos[x]++;				(12 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(13, 1);		// getfield
		code.expect(0, 2);		// s
		code.expect(2, 1);		// load_0
		code.expect(41, 1);		// dup2
		code.expect(34, 1); 	// aload
		code.expect(16, 1);		// const_1
		code.expect(23, 1);		// add
		code.expect(35, 1);		// astore
							// val.neg[-x]++;             (13 byte)
		code.expect(11, 1);  // getstatic
		code.expect( 0, 2);  // s
		code.expect(13, 1);  // getfield
		code.expect( 1, 2);  // s
		code.expect( 2, 1);  // load_0
		code.expect(28, 1);  // neg
		code.expect(41, 1);  // dup2
		code.expect(34, 1);  // aload
		code.expect(16, 1);  // const_1
		code.expect(23, 1);  // add
		code.expect(35, 1);  // astore
					// read(x);                       (2 byte)
		code.expect(53, 1);  // read
		code.expect( 7, 1);  // store_0

		Code.write(code);

		code.setCodeSize(84);
		code.setDataSize(1);
		code.setMainPC(-1);
		code.verify();
	}

	public void testAllProds () {
		initParser(	"program AllProds" + LF +
		            "  final int nPoints = 2;" + LF +
		            "  final char exclaim = '!';" + LF +
		            "" + LF +
		/*line  5*/	"  class Point {" + LF +
		            "    int x, y;" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  char[] msg;" + LF +
		/*line 10*/	"  Point[] pArr;" + LF +
		            "" + LF +
		            "{" + LF +
		            "  void printError ()" + LF +
		            "    int i;" + LF +
		/*line 15*/	"  {" + LF +
		            "    /* print error message */" + LF +
		            "    i = 0;" + LF +
		            "    while(i<len(msg)){print(msg[i]);i++;}" + LF +
		            "    print(exclaim);" + LF +
		/*line 20*/	"    do {" + LF +
		            "      i--;" + LF +
		            "      if (i < 5) break;" + LF +
		            "      print(msg[i], 2);" + LF +
		            "    } while(i > 0);" + LF +
		/*line 25*/	"  }" + LF +
		            "" + LF +
		            "  int getX (Point p) { return p.x; }" + LF +
		            "" + LF +
		            "  void setX (int x, Point p) {" + LF +
		/*line 30*/	"    if (x > -10 && x <= 9 ||" + LF +
		            "        x == 4*3/4%4+12-11*(2147483647/(-2147483647-1)%2) )" + LF +
		            "      p.x = x;" + LF +
		            "    else printError();" + LF +
		            "  }" + LF +
		/*line 35*/	"" + LF +
		            "" + LF +
		            "  void setMsg () {" + LF +
		            "    msg = new char[5];" + LF +
		            "    msg[0] = 'E';" + LF +
		/*line 40*/	"    msg[1] = 'r';" + LF +
		            "    msg[2] = 'r';" + LF +
		            "    msg[3] = 'o';" + LF +
		            "    msg[4] = 'r';" + LF +
		            "  }" + LF +
		/*line 45*/	"" + LF +
		            "  void setPoints ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		            "    pArr = new Point[nPoints];" + LF +
		/*line 50*/	"    i = 0;" + LF +
		            "    while (i < len(pArr)) {" + LF +
		            "      pArr[i] = new Point;" + LF +
		            "      pArr[i].x = i;" + LF +
		            "      pArr[i].y = i;" + LF +
		/*line 55*/	"      i++;" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  void main ()" + LF +
		/*line 60*/	"    int i;" + LF +
		            "  {" + LF +
		            "    setMsg();" + LF +
		            "    setPoints();" + LF +
		            "    read(i);" + LF +
		/*line 65*/	"    if (i < nPoints) {" + LF +
		            "      print(getX(pArr[i]));" + LF +
		            "      print(pArr[i].y);" + LF +
		            "    }" + LF +
		            "    else" + LF +
		/*line 70*/	"      printError();" + LF +
		            "    print('\\n');" + LF +
		            "  }" + LF +
		            "}");

		Parser.parse(output);
		output.verify();

		////////////////////////
		// check symbol table //
		////////////////////////

		Obj actual, expected, locals;
		Struct pointT;

		actual = Tab.find("AllProds");
		expected = new Obj(Obj.Prog, "AllProds", Tab.noType);
		// We cheat here, because we will compare all the rest later.
		expected.locals = actual.locals;
		assertEquals("Program 'AllProds'", expected, actual);

		actual = actual.locals;  // final int nPoints = 2;
		expected = new Obj(Obj.Con, "nPoints", Tab.intType);
		expected.val = 2;
		assertEquals("Constant 'nPoints'", expected, actual);

		actual = actual.next;  // final char exclaim = '!';
		expected = new Obj(Obj.Con, "exclaim", Tab.charType);
		expected.val = 33;
		assertEquals("Constant 'exclaim'", expected, actual);

		actual = actual.next;  // class Point { int x; int y; }
		pointT = new Struct(Struct.Class);
		pointT.n = 2;
		expected = new Obj(Obj.Var, "x", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Field 'x'", expected, actual.type.fields);
		pointT.fields = expected;
		expected = new Obj(Obj.Var, "y", Tab.intType);
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Field 'y'", expected, actual.type.fields.next);
		pointT.fields.next = expected;
		expected = new Obj(Obj.Type, "Point", pointT);
		assertEquals("Class 'Point'", expected, actual);

		actual = actual.next;  // char[] msg;
		expected = new Obj(Obj.Var, "msg", new Struct(Struct.Arr, Tab.charType));
		expected.adr = 0;
		expected.level = 0;
		assertEquals("Global 'msg'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Var, "pArr", new Struct(Struct.Arr, pointT));
		expected.adr = 1;
		expected.level = 0;
		assertEquals("Global 'pArr'", expected, actual);

		actual = actual.next;  // void printError () int i;
		expected = new Obj(Obj.Var, "i", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Local 'i'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Meth, "printError", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = locals;
		assertEquals("Method 'printError'", expected, actual);

		actual = actual.next;  // int getX (Point p)
		expected = new Obj(Obj.Var, "p", pointT);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Param 'p'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Meth, "getX", Tab.intType);
		expected.adr = 0;
		expected.nPars = 1;
		expected.locals = locals;
		assertEquals("Method 'getX'", expected, actual);

		actual = actual.next;  // void setX (int x, Point p)
		expected = new Obj(Obj.Var, "x", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Param 'x'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Var, "p", pointT);
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Param 'p'", expected, actual.locals.next);
		locals.next = expected;
		expected = new Obj(Obj.Meth, "setX", Tab.noType);
		expected.adr = 0;
		expected.nPars = 2;
		expected.locals = locals;
		assertEquals("Method 'setX'", expected, actual);

		actual = actual.next;  // void setMsg ()
		expected = new Obj(Obj.Meth, "setMsg", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'setMsg'", expected, actual);

		actual = actual.next;  // void setPoints () int i;
		expected = new Obj(Obj.Var, "i", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Local 'i'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Meth, "setPoints", Tab.noType);
		expected.adr = 0;
		expected.level = 0;
		expected.locals = locals;
		assertEquals("Method 'setPoints'", expected, actual);

		actual = actual.next;  // void main () int i;
		// locals are the same as above
		assertEquals("Local 'i'", locals, actual.locals);
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = locals;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);

		//////////////////////////
		// check generated code //
		//////////////////////////

			//   i = 0;                           (2 byte)
		code.expect(15, 1);  // const_0
		code.expect( 7, 1);	 // store_0
			//     print(msg[i]);                 (7 byte)
		code.expect(11, 1);  // getstatic
		code.expect( 0, 2);  // s
		code.expect( 2, 1);  // load_0
		code.expect(36, 1);  // baload
		code.expect(15, 1);  // const_0
		code.expect(56, 1);  // bprint
			//     i++;                           (3 byte)
		code.expect(31, 1);  // inc
		code.expect( 0, 1);  // b1
		code.expect( 1, 1);  // b2
			//   print(exclaim);                  (7 byte)
		code.expect(22, 1);  // const
		code.expect(33, 4);  // w
		code.expect(15, 1);  // const_0
		code.expect(56, 1);  // bprint
			//     i--;                           (3 byte)
		code.expect(31, 1);  // inc
		code.expect( 0, 1);  // b1
		code.expect(-1, 1);  // b2
			//     print(msg[i], 2);              (7 byte)
		code.expect(11, 1);  // getstatic
		code.expect( 0, 2);  // s
		code.expect( 2, 1);  // load_0
		code.expect(36, 1);  // baload
		code.expect(17, 1);  // const_2
		code.expect(56, 1);  // bprint
				// return p.x;						(6 byte)
		code.expect(2, 1);		// load_0
		code.expect(13, 1);		// getfield
		code.expect(0, 2);		// s
				//      4 * 3 / 4 % 4 + 12 -	(40 byte)
				//          11 * (2147483647 / (-2147483647 - 1) % 2))
		code.expect(19, 1);  // const_4
		code.expect(18, 1);  // const_3
		code.expect(25, 1);  // mul
		code.expect(19, 1);  // const_4
		code.expect(26, 1);  // div
		code.expect(19, 1);  // const_4
		code.expect(27, 1);  // rem
		code.expect(22, 1);  // const
		code.expect(12, 4);  // w
		code.expect(23, 1);  // add
		code.expect(22, 1);  // const
		code.expect(11, 4);  // w
		code.expect(22, 1);  // const
		code.expect(2147483647, 4);	// w
		code.expect(22, 1);  // const
		code.expect(-2147483647, 4);	// w
		code.expect(16, 1);  // const_1
		code.expect(24, 1);  // sub
		code.expect(26, 1);  // div
		code.expect(17, 1);  // const_2
		code.expect(27, 1);  // rem
		code.expect(25, 1);  // mul
		code.expect(24, 1);  // sub
					// p.x = x;						(5 byte)
		code.expect(3, 1);		// load_1
		code.expect(2, 1);		// load_0
		code.expect(14, 1);		// putfield
		code.expect(0, 2);		// s
				// msg = new char[5];				(6 byte)
		code.expect(20, 1);		// const_5
		code.expect(33, 1);		// newarray
		code.expect(0, 1);		// b
		code.expect(12, 1);		// putstatic
		code.expect(0, 2);		// s
				// msg[0] = 'E';					(10 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(15, 1);		// const_0
		code.expect(22, 1);		// const
		code.expect(69, 4);		// w
		code.expect(37, 1);		// bastore
				// msg[1] = 'r';					(10 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(16, 1);		// const_1
		code.expect(22, 1);		// const
		code.expect(114, 4);	// w
		code.expect(37, 1);		// bastore
				// msg[2] = 'r';					(10 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(17, 1);		// const_2
		code.expect(22, 1);		// const
		code.expect(114, 4);	// w
		code.expect(37, 1);		// bastore
				// msg[3] = 'o';					(10 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(18, 1);		// const_3
		code.expect(22, 1);		// const
		code.expect(111, 4);	// w
		code.expect(37, 1);		// bastore
				// msg[4] = 'r';					(10 byte)
		code.expect(11, 1);		// getstatic
		code.expect(0, 2);		// s
		code.expect(19, 1);		// const_4
		code.expect(22, 1);		// const
		code.expect(114, 4);	// w
		code.expect(37, 1);		// bastore
				// pArr = new Point[nPoints];		(6 byte)
		code.expect(17, 1);		// const_2;
		code.expect(33, 1);		// newarray
		code.expect(1, 1);		// b
		code.expect(12, 1);		// putstatic
		code.expect(1, 2);		// s
				// i = 0;							(2 byte)
		code.expect(15, 1);		// const_0;
		code.expect(7, 1);		// store_0;
					// pArr[i] = new Point;			(8 byte)
		code.expect(11, 1);		// getstatic
		code.expect(1, 2);		// s
		code.expect(2, 1);		// load_0
		code.expect(32, 1);		// new
		code.expect(8, 2);		// s
		code.expect(35, 1);		// astore
					// pArr[i].x = i;				(9 byte)
		code.expect(11, 1);		// getstatic
		code.expect(1, 2);		// s
		code.expect(2, 1);		// load_0
		code.expect(34, 1);		// aload
		code.expect(2, 1);		// load_0
		code.expect(14, 1);		// putfield
		code.expect(0, 2);		// s
					// pArr[i].y = i;				(9 byte)
		code.expect(11, 1);		// getstatic
		code.expect(1, 2);		// s
		code.expect(2, 1);		// load_0
		code.expect(34, 1);		// aload
		code.expect(2, 1);		// load_0
		code.expect(14, 1);		// putfield
		code.expect(1, 2);		// s
					// i++;							(3 byte)
		code.expect(31, 1);		// inc
		code.expect(0, 1);		// b1
		code.expect(1, 1);		// b2
				// read(i);							(2 byte)
		code.expect(53, 1);		// read
		code.expect(7, 1);		// store_0
					// print(getX(pArr[i]));		(10 byte)
		code.expect(11, 1);		// getstatic
		code.expect(1, 2);		// s
		code.expect(2, 1);		// load_0
		code.expect(15, 1);		// const_0
		code.expect(54, 1);		// print
					// print(pArr[i].y);			(10 byte)
		code.expect(11, 1);		// getstatic
		code.expect(1, 2);		// s
		code.expect(2, 1);		// load_0
		code.expect(34, 1);		// aload
		code.expect(13, 1);		// getfield
		code.expect(1, 2);		// s
		code.expect(15, 1);		// const_0
		code.expect(54, 1);		// print
				// print(eol);						(7 byte)
		code.expect(22, 1);		// const
		code.expect(10, 4);		// w
		code.expect(15, 1);		// const_0
		code.expect(56, 1);		// bprint

		Code.write(code);

		code.setCodeSize(191);
		code.setDataSize(2);
		code.setMainPC(-1);
		code.verify();
	}

	// -----------------------------------------------
	// ---------- test programs with errors ----------

	public void testScanErrIn1stSymbol () {
		initParser("");

		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 1: ",  "INVALID_CHAR", " ''");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongStart () {
		initParser("noclass WS { }");

		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 1: program expected");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testNoProgName () {
		initParser(	"program { }");

		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 9: identifier expected");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testUndefName () {
		initParser("program UN" + LF +
		           "  integer x;" + LF +
		           "{ }");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 11: integer", "NOT_FOUND");
		output.addExpectedLine("-- line 3, col 3: main", "NOT_FOUND");
		output.verify();
		assertEquals("error count", 2, Parser.Errors.count());
	}

	public void testWrongConstType () {
		initParser("program WCT" + LF +
		           "  final char ch = 32;" + LF +
		           "{ void main () { } }");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 19: ", "CONST_TYPE");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongConstDecl () {
		initParser("program WCD" + LF +
		           "  final char ch = a;" + LF +
		           "{ void main () { } }");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 19: ", "CONST_DECL");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongGlobalDecl () {
		initParser("program WGD" + LF +
		           "  final int MAX = 10;" + LF +
		           "  123, i, x;" + LF +
		           "  char c1, c2, c3;" + LF +
		           "{ void main () { } }");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 3: ", "INVALID_DECL");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());

		////////////////////////
		// check symbol table //
		////////////////////////

		Obj actual, expected;

		actual = Tab.find("WGD");
		expected = new Obj(Obj.Prog, "WGD", Tab.noType);
		// We cheat here, because we will compare all the rest later.
		expected.locals = actual.locals;
		assertEquals("Program 'WGD'", expected, actual);

		actual = actual.locals;  // final int MAX = 10;
		expected = new Obj(Obj.Con, "MAX", Tab.intType);
		expected.val = 10;
		assertEquals("Constant 'MAX'", expected, actual);

		actual = actual.next;  // char c1;
		expected = new Obj(Obj.Var, "c1", Tab.charType);
		expected.adr = 0;
		expected.level = 0;
		assertEquals("Global 'c1'", expected, actual);

		actual = actual.next;  // char c2;
		expected = new Obj(Obj.Var, "c2", Tab.charType);
		expected.adr = 1;
		expected.level = 0;
		assertEquals("Global 'c2'", expected, actual);

		actual = actual.next;  // char c3;
		expected = new Obj(Obj.Var, "c3", Tab.charType);
		expected.adr = 2;
		expected.level = 0;
		assertEquals("Global 'c3'", expected, actual);

		actual = actual.next;  // void main ()
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	public void testWrongMethDecl () {
		initParser("program WMD {" + LF +
		           "  return main ()" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 3: ", "METH_DECL");

		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongStat () {
		initParser(
				"program WS {" + LF +
				"  void main () {  " + LF +
				"    if (0 > 2) ;" + LF +
				"    123;" + LF +
				"    while (5 < 6) break;" + LF +
				"    567 no start on ident;" + LF +
				"  }" + LF +
				"}"
		);

		Parser.parse(output);
		output.addExpectedLine("-- line 4, col 5: ", "INVALID_STAT");
		output.addExpectedLine("-- line 6, col 5: ", "INVALID_STAT");
		output.verify();
		assertEquals("error count", 2, Parser.Errors.count());
	}

	public void testWrongFactor () {
		initParser("program WF {" + LF +
		           "  void main () int i; {  " + LF +
		           "    if (i > if) print('\\n');" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 13: ", "INVALID_FACT");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongRelOp () {
		initParser("program WRO {" + LF +
		           "  void main () int i; {  " + LF +
		           "    if (i x 5) print('\\n');" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 11: ", "REL_OP");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongDesignFollow () {
		initParser("program WDF {" + LF +
		           "  void main () int i; {  " + LF +
		           "    i**;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 6: ", "DESIGN_FOLLOW");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testIncompTypes () {
		initParser("program IT {" + LF +
		           "  void main () int i; {  " + LF +
		           "    i = 'a';" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 12: ", "INCOMP_TYPES");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testNoIntOps () {
		initParser("program NIO {" + LF +
		           "  void main () { } " + LF +
		           "  int calc () int i; char c; {  " + LF +
		           "    return -c;" + LF +
		           "    return i + c;" + LF +
		           "    return i / c;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 4, col 14: ", "NO_INT_OP");
		output.addExpectedLine("-- line 5, col 17: ", "NO_INT_OP");
		output.addExpectedLine("-- line 6, col 17: ", "NO_INT_OP");
		output.verify();
		assertEquals("error count", 3, Parser.Errors.count());
	}

	public void testNoInteger () {
		initParser("program NI {" + LF +
		           "  void main () char ch; {  " + LF +
		           "    ch++;" + LF +
		           "    ch--;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 7: ", "NO_INT");
		output.addExpectedLine("-- line 4, col 7: ", "NO_INT");
		output.verify();
		assertEquals("error count", 2, Parser.Errors.count());
	}

	public void testNoClassType () {
		initParser("program NCT" + LF +
		           "  class Point { int x, y; }" + LF +
		           "{ void main () Point p; {" + LF +
		           "    p = new Point;" + LF +
		           "    p = new int;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 5, col 16: ", "NO_CLASS_TYPE");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testMainVar () {
		initParser("program MV" + LF +
		           "  int main;" + LF +
		           "{ }");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 3: ", "NO_MAIN");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testMainNotVoid () {
		initParser("program MNV {" + LF +
		           "  char main () { }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 16: ", "VOID_MAIN");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testMainWithParams () {
		initParser("program MWP {" + LF +
		           "  void main (int i) { }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 21: ", "NO_PARAMS");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongReadValue () {
		initParser("program WRV {" + LF +
		           "  void main () int[] ia; { " + LF +
		           "    read(ia);" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 12: ", "READ_VALUE");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongPrintValue () {
		initParser("program WPV {" + LF +
		           "  void main () int[] ia; { " + LF +
		           "    print(ia);" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 13: ", "PRINT_VALUE");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testNoReturnVal () {
		initParser("program NRV {" + LF +
		           "  int getX () { return; }" + LF +
		           "  void main () {}" + LF +
		           "}");
		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 23: ", "NO_RETURN");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongReturnVal1 () {
		initParser("program WRV1 {" + LF +
		           "  int getX () { return 'c'; }" + LF +
		           "  void main () {}" + LF +
		           "}");
		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 27: ", "RETURN_TYPE");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testWrongReturnVal2 () {
		initParser("program WRV2 {" + LF +
		           "  void doIt () { return 5; }" + LF +
		           "  void main () {}" + LF +
		           "}");
		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 25: ", "RETURN_VOID");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testNoType () {
		initParser("program NT" + LF +
		           "  final int S = 10;" + LF +
		           "  S s;" + LF +
		           "  class Point { int x, y; }" + LF +
		           "{" + LF +
		           "  void main () Point p; {" + LF +
		           "    p = new S;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 5: ", "NO_TYPE");
		output.addExpectedLine("-- line 7, col 14: ", "NO_TYPE");
		output.verify();
		assertEquals("error count", 2, Parser.Errors.count());
	}

	public void testDoubleDecl () {
		initParser("program DD " + LF +
		           "  int x, x;" + LF +
		           "{" + LF +
		           "  void main () int o; char c, o; { }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 2, col 11: x ", "DECL_NAME");
		output.addExpectedLine("-- line 4, col 32: o ", "DECL_NAME");
		output.verify();
		assertEquals("error count", 2, Parser.Errors.count());
	}

	public void testNoField () {
		initParser("program NF" + LF +
		           "  class Point { int x, y; }" + LF +
		           "{" + LF +
		           "  void main () Point p; {" + LF +
		           "    print(p.x);" + LF +
		           "    print(p.z);" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 6, col 14: z ", "NO_FIELD");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testNoArray () {
		initParser("program NA " + LF +
		           "  int x;" + LF +
		           "{" + LF +
		           "  void main () {" + LF +
		           "    x[1]++;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 5, col 8: ", "NO_ARRAY");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testNoArraySize () {
		initParser("program NAS" + LF +
		           "  int[] ia;" + LF +
		           "{" + LF +
		           "  void main () {" + LF +
		           "    ia = new int['e'];" + LF +
		           "    ia[1] = 4;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 5, col 21: ", "ARRAY_SIZE");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testNoArrayIndex () {
		initParser("program NAI" + LF +
		           "  int[] ia;" + LF +
		           "{" + LF +
		           "  void main () {" + LF +
		           "    ia = new int[3];" + LF +
		           "    ia['a'] = 4;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 6, col 11: ", "ARRAY_INDEX");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testEqCheck () {
		initParser("program EC" + LF +
		           "  int[] arrA, arrB;" + LF +
		           "{" + LF +
		           "  void main () {" + LF +
		           "    if (arrA == arrB) print('X');" + LF +
		           "    else if (arrA > arrB) print('A');" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		// Keine Fehlermeldung bei bung 4!
		output.verify();
		assertEquals("error count", 0, Parser.Errors.count());
	}

	public void testLessParams () {
		initParser("program LP {" + LF +
		           "  void m1 (int x, int y, int z) { }" + LF +
		           "  void main () {" + LF +
		           "    m1(1,2,3);" + LF +
		           "    m1(1,2);" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		// Keine Fehlermeldung bei bung 4!
		output.verify();
		assertEquals("error count", 0, Parser.Errors.count());
	}

	public void testMoreParams () {
		initParser("program MP {" + LF +
		           "  void m1 (int x, int y, int z) { }" + LF +
		           "  void main () {" + LF +
		           "    m1(1,2,3);" + LF +
		           "    m1(1,2,3,4);" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		// Keine Fehlermeldung bei bung 4!
		output.verify();
		assertEquals("error count", 0, Parser.Errors.count());
	}

	public void testNoLoop () {
		initParser(
				"program NL {" + LF +
				"  void main () {" + LF +
				"    break;" + LF +
				"  }" + LF +
				"}"
		);

		Parser.parse(output);
		// Keine Fehlermeldung bei bung 4!
		output.verify();
		assertEquals("error count", 0, Parser.Errors.count());
	}

	public void testNoMeth () {
		initParser("program NM" + LF +
		           "  int n;" + LF +
		           "{" + LF +
		           "  void main () {" + LF +
		           "    n();" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		// Keine Fehlermeldung bei bung 4!
		output.verify();
		assertEquals("error count", 0, Parser.Errors.count());
	}

	public void testParamType () {
		initParser("program PT {" + LF +
		           "  void m (int x) { }" + LF +
		           "  void main () {" + LF +
		           "    m(1);" + LF +
		           "    m('a');" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		// Keine Fehlermeldung bei bung 4!
		output.verify();
		assertEquals("error count", 0, Parser.Errors.count());
	}

	public void testNoVar () {
		initParser("program NV {" + LF +
		           "  int m () { return 1; }" + LF +
		           "  void main () int x; {" + LF +
		           "    x = m();" + LF +
		           "    m = x;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 5, col 10: ", "NO_VAR");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	public void testNoVarInc () {
		initParser("program NVI {" + LF +
		           "  int m () { return 1; }" + LF +
		           "  void main () int x; {" + LF +
		           "    x++;" + LF +
		           "    m++;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 5, col 6: ", "NO_VAR_INC");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	/* suggested by KBP */
	public void testResetErrDist () {
		initParser("program RED {" + LF +
		           "  void main () {" + LF +
							 "    if () { i = i + i + i; }" + LF +
							 "  }" + LF +
							 "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 9: ", "INVALID_FACT");
		output.addExpectedLine("-- line 3, col 15: i", "NOT_FOUND");
		output.verify();
		assertEquals("error count", 2, Parser.Errors.count());
	}

	public void testTypeCompatibility () {
		initParser("program TC " + LF +
		           "  class C1 { int x, y; }" + LF +
		           "  class C2 { int x, y; }" + LF +
				   "{" + LF +
				   "  C1 test () C1 c1; C2 c2; {" + LF + 
				   "    c1 = new C1;" + LF +
				   "    c2 = c1;" + LF +
				   "    return c2;" + LF +
				   "  }" + LF +
		           "  void main () {" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 7, col 12: ", "INCOMP_TYPES");
		output.addExpectedLine("-- line 8, col 14: ", "RETURN_TYPE");
		output.verify();
		assertEquals("error count", 2, Parser.Errors.count());
	}

	public void testAllProdsWithErrors () {
		initParser(	"program AllProdsWithErros" + LF +
		            "  final int nPoints = 2;" + LF +
		            "  final exclaim = '!';" + LF +
		            "" + LF +
		/*line  5*/ "  class {" + LF +
		            "    int x, y;" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  char[ msg;" + LF +
		/*line 10*/ "  Point[] pArr;" + LF +
		            "" + LF +
		            "{" + LF +
		            "  void printError ()" + LF +
		            "    int i;" + LF +
		/*line 15*/ "  {" + LF +
		            "    /* print error message */" + LF +
		            "    i = 0;" + LF +
		            "    while(i<len(msg)){print(msg[i]);i++;}" + LF +
		            "    print(exclaim);" + LF +
		/*line 20*/ "    do {" + LF +
		            "      i--;" + LF +
		            "      if (i >= 5) break;" + LF +
		            "      print(msg[i]);" + LF +
		            "    } while(i > 0);" + LF +
		/*line 25*/ "  }" + LF +
		            "" + LF +
		            "  int getX (Point p) { return p.x; }" + LF +
		            "" + LF +
		            "  void setX (int x, Point p) {" + LF +
		/*line 30*/ "    if (x > -10 && x <= 9 ||" + LF +
		            "        x == 4*3/4%4+12-11*(2147483647/(-2147483647-1)%2) )" + LF +
		            "      p.x = x;" + LF +
		            "    else printError();" + LF +
		            "  }" + LF +
		/*line 35*/ "" + LF +
		            "" + LF +
		            "  void setMsg () {" + LF +
		            "    msg = new char[6];" + LF +
		            "    msg[0] 'E';" + LF +
		/*line 40*/ "    msg[1] = 'r';" + LF +
		            "    msg[2] = 'r';" + LF +
		            "    msg[3] = 'o';" + LF +
		            "    msg[4] = 'r';" + LF +
		            "    msg[5] = '\\n';" + LF +
		/*line 45*/ "  }" + LF +
		            "" + LF +
		            "  void setPoints ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		/*line 50*/ "    pArr = new Point[nPoints];" + LF +
		            "    i = 0;" + LF +
		            "    while (i < len(pArr)) {" + LF +
		            "      pArr[i] = new Point;" + LF +
		            "      pArr[i].x = i;" + LF +
		/*line 55*/ "      pArr[i].y = i" + LF +
		            "      i++;" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "" + LF +
		/*line 60*/ "  int main ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		            "    setMsg();" + LF +
		            "    setPoints();" + LF +
		/*line 65*/ "    read(i);" + LF +
		            "    if (i < nPoints) {" + LF +
		            "      print(getX(pArr[i]));" + LF +
		            "      print(pArr[i].y);" + LF +
		            "    }" + LF +
		/*line 70*/ "    else" + LF +
		            "      printError();" + LF +
		            "    print('\\n');" + LF +
		            "  }" + LF +
		            "}");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 17: exclaim", "NOT_FOUND");
		output.addExpectedLine("-- line 5, col 9: identifier expected");
		output.addExpectedLine("-- line 9, col 9: ] expected");
		output.addExpectedLine("-- line 10, col 8: Point", "NOT_FOUND");
		output.addExpectedLine("-- line 19, col 18: ", "PRINT_VALUE");
		output.addExpectedLine("-- line 27, col 19: Point", "NOT_FOUND");
		output.addExpectedLine("-- line 27, col 34: ", "NO_CLASS");
		output.addExpectedLine("-- line 29, col 27: Point", "NOT_FOUND");
		output.addExpectedLine("-- line 32, col 11: ", "NO_CLASS");
		output.addExpectedLine("-- line 39, col 12: assignment, method call, increment (++) or decrement (--) expected");
		output.addExpectedLine("-- line 50, col 21: Point", "NOT_FOUND");
		output.addExpectedLine("-- line 53, col 26: Point", "NOT_FOUND");
		output.addExpectedLine("-- line 54, col 17: ", "NO_CLASS");
		output.addExpectedLine("-- line 55, col 17: ", "NO_CLASS");
		output.addExpectedLine("-- line 61, col 5: ", "VOID_MAIN");
		output.addExpectedLine("-- line 68, col 22: ", "NO_CLASS");
		output.verify();
		assertEquals("error count", 16, Parser.Errors.count());
	}

	public void testNoEof () {
		initParser( "program X { void main () {} } more text");

		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 31: ", "NO_EOF");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());
	}

	// -----------------------------------------------

	public static void main (String[] args) {
		junit.textui.TestRunner.run(ParserTest.class);
	}
}
